/*
The software is provided by the Institute of Commercial Cryptography Standards 
(ICCS), and is used for algorithm submissions in the Next-generation Commercial 
Cryptographic Algorithms Program. 

ICCS doesn't represent or warrant that the operation of the software will be 
uninterrupted or error-free in all cases. ICCS will take no responsibility for 
the use of the software or the results thereof, if the software is used for any 
other purposes.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "drng.h"
#include "KEM_AlgorithmInstance.h"

#define SEED_LENGTH_BYTE 64

#define KAT_KEM_SUCCESS 0
#define KAT_KEM_FOPEN_ERROR -1
#define KAT_KEM_CRYPTO_FAILURE -3

void fprintstr(FILE *file_output, unsigned char *msg, unsigned long long len);

// DRNG_ctx for generating pseudorandom numbers within the KEM scheme
DRNG_ctx drng_algorithm;

/****************output KAT_KEM_AlgorithmInstance.txt****************/
int main()
{
	unsigned char *nonce;
	// DRNG_ctx for generating seed
	DRNG_ctx drng_seed;
	char algoname_output[128];
	FILE *file_output;
	unsigned char *seed, *ss, *ss1, *ct, *pk, *sk;
	unsigned long long pk_byts, sk_byts, ss_byts, ct_byts;
	int rtn;
	// init drng_seed using nonce
	nonce = (unsigned char *)malloc(SEED_LENGTH_BYTE);
	for (int i = 0; i < SEED_LENGTH_BYTE / 4; i++)
	{
		memcpy(nonce + 4 * i, "seed", 4);
	}
	init_random_number(&drng_seed, nonce, SEED_LENGTH_BYTE);
	// open KAT_KEM_AlgorithmInstance.txt
	sprintf(algoname_output, "KAT_KEM_%s.txt", ALGORITHM_INSTANCE);
	if ((file_output = fopen(algoname_output, "wb")) == NULL)
		return KAT_KEM_FOPEN_ERROR;

	pk_byts = kem_get_pk_byts();
	sk_byts = kem_get_sk_byts();
	ss_byts = kem_get_ss_byts();
	ct_byts = kem_get_ct_byts();
	pk = (unsigned char *)calloc(pk_byts, sizeof(unsigned char));
	sk = (unsigned char *)calloc(sk_byts, sizeof(unsigned char));
	ss = (unsigned char *)calloc(ss_byts, sizeof(unsigned char));
	ss1 = (unsigned char *)calloc(ss_byts, sizeof(unsigned char));
	ct = (unsigned char *)calloc(ct_byts, sizeof(unsigned char));
	seed = (unsigned char *)calloc(SEED_LENGTH_BYTE, sizeof(unsigned char));
	for (int i = 0; i < 10; i++)
	{
		fprintf(file_output, "count = %d\n", i);
		// generate seed using drng_seed
		get_random_number(&drng_seed, seed, SEED_LENGTH_BYTE * 8);
		// print seed
		fprintf(file_output, "seedlen = %08X\n", SEED_LENGTH_BYTE);
		fprintf(file_output, "seed = ");
		fprintstr(file_output, seed, SEED_LENGTH_BYTE);
		fprintf(file_output, "\n");

		/****************the KEM scheme starts here****************/
		// init drng_algorithm using seed
		init_random_number(&drng_algorithm, seed, SEED_LENGTH_BYTE);

		/****************       Key generate       ****************/
		rtn = kem_keygen(pk, &pk_byts, sk, &sk_byts);
		if (rtn)
		{
			printf("kem_keygen returned %d\n", rtn);
			return KAT_KEM_CRYPTO_FAILURE;
		}
		// print public key
		if (OUTPUT_BLANK_TEST_VECTORS)
			fprintf(file_output, "pklen = \n");
		else
			fprintf(file_output, "pklen = %08llX\n", pk_byts);
		fprintf(file_output, "pk = ");
		fprintstr(file_output, pk, pk_byts);
		fprintf(file_output, "\n");
		// print private key
		if (OUTPUT_BLANK_TEST_VECTORS)
			fprintf(file_output, "sklen = \n");
		else
			fprintf(file_output, "sklen = %08llX\n", sk_byts);
		fprintf(file_output, "sk = ");
		fprintstr(file_output, sk, sk_byts);
		fprintf(file_output, "\n");
		/****************        Encapsulate       ****************/
		rtn = kem_enc(pk, pk_byts, ss, &ss_byts, ct, &ct_byts);
		if (rtn)
		{
			printf("kem_enc returned %d\n", rtn);
			return KAT_KEM_CRYPTO_FAILURE;
		}
		// print ciphertext
		if (OUTPUT_BLANK_TEST_VECTORS)
			fprintf(file_output, "ctlen = \n");
		else
			fprintf(file_output, "ctlen = %08llX\n", ct_byts);
		fprintf(file_output, "ct = ");
		fprintstr(file_output, ct, ct_byts);
		fprintf(file_output, "\n");
		// print encapsulation shared secret key
		if (OUTPUT_BLANK_TEST_VECTORS)
			fprintf(file_output, "sslen = \n");
		else
			fprintf(file_output, "sslen = %08llX\n", ss_byts);
		fprintf(file_output, "ss = ");
		fprintstr(file_output, ss, ss_byts);
		fprintf(file_output, "\n");

		/****************        Decapsulate       ****************/
		rtn = kem_dec(sk, sk_byts, ct, ct_byts, ss1, &ss_byts);
		if (rtn)
		{
			printf("kem_dec returned %d\n", rtn);
			return KAT_KEM_CRYPTO_FAILURE;
		}
		// verify decapsulation result
		if (memcmp(ss, ss1, ss_byts))
		{
			printf("decapsulation shared secret key != encapsulation shared secret key\n");
			return KAT_KEM_CRYPTO_FAILURE;
		}
		fprintf(file_output, "\n");
	}
	free(seed);
	free(ct);
	free(ss);
	free(sk);
	free(pk);
	free(nonce);
	// close KAT_KEM_AlgorithmInstance.txt
	fclose(file_output);
	return KAT_KEM_SUCCESS;
}

void fprintstr(FILE *file_output, unsigned char *msg, unsigned long long len)
{
	for (unsigned long long i = 0; i < len; i++)
		fprintf(file_output, "%02X", msg[i]);
}